package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.Sequence.Position;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.lang.ref.*;

/** a traversal of the grammar performed by mapping from Elements to a lattice and computing the resulting LUB */
abstract class Walk<T> {
    protected HashSet<Element> acc = new HashSet<Element>();
    protected abstract T bottom(Element e);

    protected final Cache c;
    public Walk() { this(null); }
    public Walk(Cache c) { this.c = c; }

    public  T sequence(Sequence s) {
        T ret = bottom(s);
        for(Position p = s.firstp(); p!=null && !p.isLast(); p = p.next())
            ret = sequence(s, ret, walk(p.element()));
        return ret;
    }
    
    public      T walkAtom(Atom r) { return walk(r); }
    public      T union(Union u, T a, T b) { return  bottom(u); }
    public      T sequence(Sequence s, T a, T b) { return  bottom(s); }
    protected   T walk(Element e) {
        if (acc.contains(e)) return bottom(e);
        acc.add(e);
        return walk2(e);
    }

    protected T walk2(Element e) {
        if      (e instanceof Atom)     return walkAtom((Atom)e);
        else if (e instanceof Sequence) return sequence((Sequence)e);
        else if (e instanceof Union) {
            T ret = bottom(e);
            for(Sequence s : (Union)e) {
                ret = union((Union)e, ret, walk(s));

                // FIXME
                for(Sequence ss : s.needs()) ret = union((Union)e, ret, walk(ss));
                for(Sequence ss : s.hates()) ret = union((Union)e, ret, walk(ss));
            }
            return ret;
        } else {
            throw new Error("unknown element of class " + e.getClass().getName() + ": " + e);
        }
    }

    static class YieldSet extends Walk<HashSet<Element>> {
        private final Element e;
        public final HashSet<Element> walk() { return walk(e); }
        public YieldSet(Element e, Cache c)  { super(c); this.e = e; }
        public HashSet<Element> bottom(Element e)     { return acc; }
        public HashSet<Element> sequence(Sequence seq) { return bottom(seq); }
        public HashSet<Element> walkAtom(Atom r) {
            c.atoms.put(e, c.atoms.get(e)==null ? r : c.atoms.get(e).union(r));
            return super.walkAtom(r);
        }
    }


    // Boolean //////////////////////////////////////////////////////////////////////////////

    static class PossiblyEpsilon extends Walk<Boolean> {
        public Boolean walkAtom(Atom r) { return false; }
        public Boolean sequence(Sequence s, Boolean a, Boolean b)  { return new Boolean(a && b); }
        public Boolean union(Union u, Boolean a, Boolean b)     { return new Boolean(a || b); }
        public Boolean bottom(Element e)    { return (e instanceof Union) ? false : true; }
        private HashMap<Element,Boolean> hm = new HashMap<Element,Boolean>();
        protected Boolean walk(Element e) {
            if (hm.get(e) != null) return hm.get(e);
            hm.put(e, false);
            Boolean ret = walk2(e);
            hm.put(e, ret);
            return ret;
        }
    }


    // Input-Set //////////////////////////////////////////////////////////////////////////////

    static abstract class WalkTokenSet<Tok extends Input> extends Walk<Topology<Tok>> {
        public Topology<Tok> cs;
        public WalkTokenSet(Topology<Tok> cs)          { this.cs = cs; }
        public WalkTokenSet(Topology<Tok> cs, Cache c) { super(c); this.cs = cs; }
        public Topology<Tok> bottom(Element e)         { return cs; }
        public Topology<Tok> walkAtom(Atom r)          { cs = cs.union(r.getTokenTopology()); return cs; }
    }

    static class First<Tok extends Input> extends WalkTokenSet<Tok> {
        public First(Topology<Tok> cs, Walk.Cache cache) { super(cs, cache); }
        public Topology<Tok> sequence(Sequence seq) {
            for(Position p = seq.firstp(); p!=null && !p.isLast(); p = p.next()) {
                walk(p.element());
                if (!c.possiblyEpsilon(p.element())) break;
            }
            return cs;
        }
    }

    static class Follow<Tok extends Input> extends WalkTokenSet<Tok> {
        private final Element me;
        private final HashSet<Element> all;
        private boolean eof = false;
        public boolean includesEof() { return eof; }
        public Follow(Topology<Tok> cs, Element me, HashSet<Element> all, Cache c)  { super(cs, c); this.me = me; this.all = all; }
        public Topology<Tok> bottom(Element e)                       { return cs; }
        public Topology<Tok> sequence(Sequence seq)                  { return cs; }
        public Topology<Tok> walkAtom(Atom r) { return walk((Element)r); }
        public Topology<Tok> walk(Element e) {
            if (acc.contains(e)) return bottom(e);
            acc.add(e);

            if (c != null) {
                Topology<Tok> cached = (Topology<Tok>)c.follow.get(e);
                if (cached != null) {
                    cs = cs.union(cached);
                    eof |= c.eof.get(e);
                    return cs;
                }
            }

            Topology<Tok> cso = cs;
            boolean eofo = eof;
            eof = c.eof.get(e) != null && c.eof.get(e).booleanValue();
            cs = cso.empty();

            for(Element x : all) {
                boolean matched = false;
                if (!(x instanceof Sequence)) continue;
                Sequence a = (Sequence)x;
                Position mp = null;
                for(Position pos = a.firstp(); pos != null && !pos.isLast(); pos = pos.next()) {
                    if (matched) cs = cs.union(c.first(pos.element(), cs.empty()));
                    if (pos.isLast()) { matched = (matched && c.possiblyEpsilon(pos.element())); continue; }
                    boolean good = false;
                    if (e instanceof Atom) {
                        Topology top = c.atoms.get(pos.element());
                        if (top==null) continue;
                        if (!(top.containsAll(((Atom)e)))) continue;
                    } else {
                        if (c.ys.contains(pos.element(),e)) good = true;
                    }
                    if (good) {
                        mp = pos;
                        matched = true;
                    }
                }
                if (matched) walk(a);
            }

            if (e instanceof Sequence) {
                Sequence s = (Sequence)e;
                if (s.follow != null) cs = cs.intersect(s.follow.getTokenTopology());
            }

            if (c != null && e==me) {
                c.follow.put(e, cs);
                c.eof.put(e, eof);
            }

            cso = cso.union(cs);
            cs = cso;
            eofo |= eof;
            eof = eofo;

            return cs;
        }
    }

    static class Cache {
        public final HashMap<Element,Boolean> possiblyEpsilon = new HashMap<Element,Boolean>();
        public HashMap<Element,Boolean> eof = new HashMap<Element,Boolean>();
        public HashMap<Element,Topology> follow = new HashMap<Element,Topology>();
        public HashMapBag<Element,Element>  ys = new HashMapBag<Element,Element>();
        public HashMap<Element,Topology> atoms = new HashMap<Element,Topology>();
        public <Tok extends Input> Topology<Tok> first(Element e, Topology<Tok> empty) {
            return new Walk.First<Tok>(empty, this).walk(e);
        }
        final boolean possiblyEpsilon(Element e) {
            Walk.Cache cache = this;
            Boolean ret = possiblyEpsilon.get(e);
            if (ret != null) return ret.booleanValue();
            ret = new Walk.PossiblyEpsilon().walk(e) ? Boolean.TRUE : Boolean.FALSE;
            possiblyEpsilon.put(e, ret);
            return ret;
        }
    }
}
